/*
 * (c) 2023, SWD Embedded Systems Limited, http://www.kpda.ru
 */

#include <errno.h>

#include <stdint.h>
#include <inttypes.h>

#include <io-pkt/iopkt_driver.h>
#include <sys/io-pkt.h>
#include <sys/syspage.h>
#include <sys/device.h>
#include <device_qnx.h>
#include <net/if_ether.h>
#include <net/if_media.h>
#include <net/netbyte.h>
#include <net80211/ieee80211_var.h>
#include <netdrvr/mdi.h>
#include <netdrvr/nicsupport.h>
#include <net/ifdrvcom.h>
#include <sys/sockio.h>

#include <sys/slog.h>
#include <sys/slogcodes.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#ifdef VARIANT_a
#include <nw_dl.h>
#endif

#include "sun8i.h"

#ifdef VARIANT_a
struct nw_dll_syms sun8i_syms[] = {
        {"iopkt_drvr_entry", &IOPKT_DRVR_ENTRY_SYM(sam)},
        {NULL, NULL}
};
#endif

int sun8i_entry(void *dll_hdl, struct _iopkt_self *iopkt, char *options);
int sun8i_attach(struct device *, struct device *, void *);
int sun8i_init(struct ifnet *);
void sun8i_stop(struct ifnet *, int);
int sun8i_ioctl(struct ifnet *, unsigned long, caddr_t);
void sun8i_start(struct ifnet *);
const struct sigevent * sun8i_isr(void *, int);
int sun8i_process_interrupt(void *, struct nw_work_thread *);
int sun8i_enable_interrupt(void *);
void sun8i_shutdown(void *);
short unsigned int sun8i_mdio_read(void *arg, unsigned char phy_addr, unsigned char phy_reg);
void sun8i_mdio_write(void *arg, unsigned char phy_addr, unsigned char phy_reg, short unsigned int phy_data);
int sun8i_detach(struct device *, int);
int sun8i_disable_interrupt(void *arg);
void bsd_mii_initmedia(sun8i_dev_t *sun8i);
void bsd_mii_finimedia(sun8i_dev_t *sun8i);

struct _iopkt_drvr_entry IOPKT_DRVR_ENTRY_SYM(sam) = IOPKT_DRVR_ENTRY_SYM_INIT(sun8i_entry);

CFATTACH_DECL(sun8i,
	sizeof(sun8i_dev_t),
	NULL,
	sun8i_attach,
	sun8i_detach,
	NULL);

static char *sun8i_opts[] = {
	NULL
};

int sun8i_parse_options(sun8i_dev_t *sun8i, const char *optstring, nic_config_t *cfg)
{
	int				opt = 0;
	int				invalid = 0;
	int				rc = EOK;
	char				*value = NULL;
	char				*options = NULL;
	char				*freeptr = NULL;
	char				*c = NULL;

	if (optstring == NULL)
	{
		return 0;
	}

	options = malloc(strlen(optstring) + 1, M_TEMP, M_NOWAIT);
	if (options == NULL)
	{
		return ENOMEM;
	}
	strcpy(options, optstring);
	freeptr = options;

	while (options && *options != '\0')
	{
		c = options;
		invalid = 0;
		if ((opt = getsubopt(&options, sun8i_opts, &value)) != -1)
		{
			switch (opt)
			{
				default:
					rc = EINVAL;
					invalid = 1;

					break;
			}

			if (sun8i == NULL)
			{
				continue;
			}
		}
		else
		{
			if (nic_parse_options(cfg, value) != EOK)
			{
					rc = EINVAL;
					invalid = 1;
			}
		}

		if (invalid)
		{
			slogf(_SLOGC_NETWORK, _SLOG_WARNING, "devnp-sun8i: unknown option %s", c);
		}
	}

	free(freeptr, M_TEMP);

	return rc;
}

// config mirror port 1 add source ports 3 both; enable mirror
// config serial_port auto_logout never
// config ports 24 state disable
// setenv serverip 10.20.42.120; setenv ipaddr 10.20.42.91; setenv gatewayip 10.20.42.1; tftpboot 0x41000000 ifs-h3-orangepi.bin; go 0x41000000
int sun8i_entry(void *dll_hdl, struct _iopkt_self *iopkt, char *options)
{
	uint32_t				val = 0;
	struct sun8i_attach_args		sun8i_args;
	struct device           		*dev = NULL;

	memset(&sun8i_args, 0, sizeof(struct sun8i_attach_args));

	sun8i_args.variant = H3_EMAC;
	sun8i_args.use_internal_phy = 1;
	sun8i_args.phy_addr = 1;
	sun8i_args.mii_mode = PHY_INTERFACE_MODE_MII;
	sun8i_args.iopkt = iopkt;
	sun8i_args.options = options;

	// Mmap System Config, Clock Control Unit, Emac address spaces
	sun8i_args.syscon_base = mmap_device_io(AW_H3_SYS_CTRL_SIZE, AW_H3_SYS_CTRL_BASE);
	if (sun8i_args.syscon_base == (uintptr_t)MAP_DEVICE_FAILED)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Problem mapping syscon memories ", __func__, __LINE__);
		return (errno);
	}

	sun8i_args.clk_base = mmap_device_io(AW_H3_CCU_SIZE, AW_H3_CCU_BASE);
	if (sun8i_args.clk_base == (uintptr_t)MAP_DEVICE_FAILED)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Problem mapping ccu memories", __func__, __LINE__);
		return (errno);
	}

	sun8i_args.emac_base = mmap_device_io(SUN8I_EMAC_SIZE, SUN8I_EMAC_BASE);
	if (sun8i_args.emac_base == (uintptr_t)MAP_DEVICE_FAILED)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Problem mapping emac memories", __func__, __LINE__);
		return (errno);
	}

	// Set up clocks (see uboot driver, dts/dtsi files)
	val = in32(sun8i_args.clk_base + BUS_CLK_GATING_REG0);
	val |= BUS_CLK_GATING_REG0_EMAC_GATING;
	out32(sun8i_args.clk_base + BUS_CLK_GATING_REG0, val);

	val = in32(sun8i_args.clk_base + BUS_SOFT_RST_REG0);
	val |= BUS_SOFT_RST_REG0_EMAC_RST;
	out32(sun8i_args.clk_base + BUS_SOFT_RST_REG0, val);

	val = in32(sun8i_args.clk_base + BUS_CLK_GATING_REG4);
	val |= BUS_CLK_GATING_REG4_EPHY_GATING;
	out32(sun8i_args.clk_base + BUS_CLK_GATING_REG4, val);

	val = in32(sun8i_args.clk_base + BUS_SOFT_RST_REG2);
	val |= BUS_SOFT_RST_REG2_EPHY_RST;
	out32(sun8i_args.clk_base + BUS_SOFT_RST_REG2, val);

	// Set up system configuration for EMAC
	val = in32(sun8i_args.syscon_base + EMAC_CLK_REG);
	if (sun8i_args.use_internal_phy)
	{
		val &= ~H3_EPHY_DEFAULT_MASK;
		val |=  H3_EPHY_DEFAULT_VALUE;
		val |= sun8i_args.phy_addr << H3_EPHY_ADDR_SHIFT;
		val &= ~H3_EPHY_SHUTDOWN;
		val |= H3_EPHY_SELECT;
		val |= H3_EPHY_LED_POL;
		val |= H3_EPHY_CLK_SEL;
	}
	else
	{
		val &= ~H3_EPHY_SELECT;
	}
	val &= ~(SC_ETCS_MASK | SC_EPIT);
	val &= ~SC_RMII_EN;
	out32(sun8i_args.syscon_base + EMAC_CLK_REG, val);

	if (dev_attach("sun", NULL, &sun8i_ca, &sun8i_args, NULL, &dev, NULL) != EOK)
	{
		return ENODEV;
	}
	dev->dv_dll_hdl = dll_hdl;

	return EOK;
}

static int sun8i_write_mac_addr(sun8i_dev_t *sun8i, uint8_t *mac)
{
	uint32_t				mac_lo = 0;
	uint32_t				mac_hi = 0;

	mac_lo = mac[0] + (mac[1] << 8) + (mac[2] << 16) + (mac[3] << 24);
	mac_hi = mac[4] + (mac[5] << 8);

	out32(sun8i->emac_base + EMAC_ADDR0_HIGH, mac_hi);
	out32(sun8i->emac_base + EMAC_ADDR0_LOW, mac_lo);

	return 0;
}

void sun8i_mdi_callback(void *handle, uint8_t phyaddr, uint8_t linkstate)
{
	int					i = 0;
	int					mode = -1;
	uint32_t				val = 0;
	char					*s = 0;
	sun8i_dev_t				*sun8i = handle;
	struct ifnet				*ifp = &sun8i->ec.ec_if;
	nic_config_t				*cfg = &sun8i->cfg;

	slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d phyaddr = %d linkstate = %d devidx = %d", __func__, __LINE__, phyaddr, linkstate, cfg->device_index);

	switch (linkstate)
	{
		case    MDI_LINK_UP:
			if ((i = MDI_GetActiveMedia(sun8i->mdi, cfg->phy_addr, &mode)) != MDI_LINK_UP)
			{
				slogf(_SLOGC_NETWORK, _SLOG_INFO, "MDI_GetActiveMedia returned %x phy_addr: %d", i, cfg->phy_addr);
				mode = 0;
			}

			switch (mode)
			{
				case MDI_10bTFD:
					cfg->media_rate = 10000L;

					break;
				case MDI_10bT:
					s = "10 BaseT Half Duplex";
					cfg->duplex = 0;
					cfg->media_rate = 10000L;

					break;
				case MDI_100bTFD:
					s = "100 BaseT Full Duplex";
					cfg->duplex = 1;
					cfg->media_rate = 100000L;

					break;
				case MDI_100bT:
					s = "100 BaseT Half Duplex";
					cfg->duplex = 0;
					cfg->media_rate = 100000L;

					break;
				case MDI_100bT4:
					s = "100 BaseT4";
					cfg->duplex = 0;
					cfg->media_rate = 100000L;

					break;
				case MDI_100bT2FD:
					s = "100 BaseT2 Full Duplex";
					cfg->duplex = 1;
					cfg->media_rate = 100000L;

					break;
				default:
					s = "Unknown";
					cfg->duplex = 0;
					cfg->media_rate = 0L;

					break;
			}
			slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d - Link up (%s) if=%d mode = %d", __func__, __LINE__, s, cfg->device_index, mode);

			val = (sun8i->emac_base + EMAC_CTL0);
			if (cfg->duplex)
			{
				val |= EMAC_CTL0_FULL_DUPLEX;
			}
			else
			{
				val &= ~EMAC_CTL0_FULL_DUPLEX;
			}

			if (cfg->media_rate == 100000L)
			{
				val |= EMAC_CTL0_100;
			}
			else
			{
				val &= ~EMAC_CTL0_100;
			}

			out32(sun8i->emac_base + EMAC_CTL0, val);

			cfg->flags &= ~NIC_FLAG_LINK_DOWN;
			sun8i->linkup = 1;
			if_link_state_change(ifp, LINK_STATE_UP);

			break;
		case    MDI_LINK_DOWN:
			cfg->media_rate = cfg->duplex = -1;
			MDI_AutoNegotiate(sun8i->mdi, cfg->phy_addr, NoWait);
			cfg->flags |= NIC_FLAG_LINK_DOWN;
			sun8i->linkup = 0;

			out32(sun8i->emac_base + EMAC_CTL0, 0);

			slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d: Link down", __func__, __LINE__);

			if_link_state_change(ifp, LINK_STATE_DOWN);

			break;
		default:
			slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d: Unknown link state %hhu", __func__, __LINE__, linkstate);

			break;
	}

}

void sun8i_MDI_MonitorPhy(void *arg)
{
	sun8i_dev_t *sun8i = arg;

	if ((sun8i->cfg.flags & NIC_FLAG_LINK_DOWN) || (sun8i->cfg.media_rate <= 0))
	{
		MDI_MonitorPhy(sun8i->mdi);
	}

	callout_msec(&sun8i->mii_callout, 3 * 1000, sun8i_MDI_MonitorPhy, sun8i);
}

void fill_random_mac(uint8_t *mac_addr)
{
	int					i = 0;
	uint64_t				clc = ClockCycles();
	uint32_t				rnd = ((clc >> 32) & 0xFFFFFF) ^ (clc & 0xFFFFFF);

	for (i = 0; i < 6; i++)
	{
		rnd = (0x9f645e7c * rnd + 0xf0176520);
		mac_addr[i] = rnd >> 24;
	}
	mac_addr[0] &= ~1;
	mac_addr[0] |= 2;
}

int sun8i_attach(struct device *parent __attribute__((unused)), struct device *self, void *aux)
{
	int					rc = 0;
	int					err = 0;
	uint8_t					mac_addr[ETHER_ADDR_LEN] = {0};
	uint32_t				val = 0;
	char					*options = NULL;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)self;
	struct sun8i_attach_args		*sun8i_args =  (struct sun8i_attach_args *)aux;
	struct ifnet				*ifp = &sun8i->ec.ec_if;
	struct _iopkt_self			*iopkt = sun8i_args->iopkt;
	int					timeout = EMAC_SOFT_RESET_TIMEOUT_USEC;

	sun8i->syscon_base = sun8i_args->syscon_base;
	sun8i->clk_base = sun8i_args->clk_base;
	sun8i->emac_base = sun8i_args->emac_base;

	options = sun8i_args->options;

	sun8i->iopkt = iopkt;
	ifp->if_softc = sun8i;

	memset(&sun8i->cfg, 0, sizeof(nic_config_t));

	// Set up requir stats for nicinfo
	sun8i->stats.un.estats.valid_stats =
		NIC_ETHER_STAT_SINGLE_COLLISIONS |
		NIC_ETHER_STAT_MULTI_COLLISIONS |
		NIC_ETHER_STAT_TX_DEFERRED |
		NIC_ETHER_STAT_XCOLL_ABORTED |
		NIC_ETHER_STAT_LATE_COLLISIONS |
		NIC_ETHER_STAT_ALIGN_ERRORS |
		NIC_ETHER_STAT_FCS_ERRORS |
		NIC_ETHER_STAT_OVERSIZED_PACKETS |
		NIC_ETHER_STAT_SYMBOL_ERRORS |
		NIC_ETHER_STAT_JABBER_DETECTED |
		NIC_ETHER_STAT_SHORT_PACKETS |
		NIC_ETHER_STAT_SQE_ERRORS |
		NIC_ETHER_STAT_INTERNAL_RX_ERRORS;

	sun8i->stats.valid_stats =
		NIC_STAT_TX_FAILED_ALLOCS | NIC_STAT_RX_FAILED_ALLOCS |
		NIC_STAT_TXED_MULTICAST | NIC_STAT_TXED_BROADCAST |
		NIC_STAT_RXED_MULTICAST | NIC_STAT_RXED_BROADCAST;

	// Set up io-pkt API (cfg - ifconfig, if_*, inter* - callouts)
	sun8i->cfg.priority = IRUPT_PRIO_DEFAULT;
	sun8i->cfg.lan = -1;
	sun8i->cfg.media_rate = -1;
	sun8i->cfg.duplex = -1;
	sun8i->cfg.mtu = ETHERMTU;
	sun8i->cfg.mru = ETHERMTU;
	sun8i->cfg.flags |= NIC_FLAG_MULTICAST;
	sun8i->cfg.phy_addr = sun8i_args->phy_addr;

	strcpy(ifp->if_xname, sun8i->dev.dv_xname);
	strcpy((char *) sun8i->cfg.uptype, "en");
	strcpy((char *) sun8i->cfg.device_description, "sun8i");

	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_ioctl = sun8i_ioctl;
	ifp->if_start = sun8i_start;
	ifp->if_init  = sun8i_init;
	ifp->if_stop  = sun8i_stop;
	IFQ_SET_READY(&ifp->if_snd);

	sun8i->inter.func = sun8i_process_interrupt;
	sun8i->inter.enable = sun8i_enable_interrupt;
	sun8i->inter.arg = sun8i;
	sun8i->iid = -1;

	sun8i->cfg.flags |= NIC_FLAG_LINK_DOWN;

	err = sun8i_parse_options(sun8i, options, &sun8i->cfg);
	if (err != EOK)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "devnp-sun8i: error parsing options");
		return err;
	}
	if (!sun8i->cfg.num_irqs)
	{
		sun8i->cfg.irq[0] = SUN8I_H3_INTERRUPT;
		++sun8i->cfg.num_irqs;
	}

	val = in32(sun8i->emac_base + EMAC_CTL1);

	// Soft reset EMAC
	if (!(val & EMAC_CTL1_SOFT_RST))
	{
		in32(sun8i->emac_base + EMAC_CTL1);
		val |= EMAC_CTL1_SOFT_RST;
		out32(sun8i->emac_base + EMAC_CTL1, val);

		do
		{
			mem_barrier();
			val = in32(sun8i->emac_base + EMAC_CTL1);
			nanospin_ns(1000);
		} while ((val & EMAC_CTL1_SOFT_RST) != 0 &&  (--timeout));

		if (!timeout)
		{
			val = in32(sun8i->emac_base + EMAC_CTL1);
			slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d Timeout EMAC_CTL1 = %#x\n", __func__, __LINE__, val);
			return -1;
		}
	}

	// Set up MDI API
	err = MDI_Register_Extended(sun8i, sun8i_mdio_write, sun8i_mdio_read, sun8i_mdi_callback, (mdi_t **)&sun8i->mdi, NULL, 0, 0);
	if (err != MDI_SUCCESS)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d - MDI_Register_Extended failed rc=%d", __func__, __LINE__, err);
		return -1;
	}
	callout_init(&sun8i->mii_callout);

	rc = MDI_InitPhy(sun8i->mdi, sun8i->cfg.phy_addr);
	if (rc != MDI_SUCCESS)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "InitPhy failed");
		return (-1);
	}
	else
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "InitPhy success");
	}

	bsd_mii_initmedia(sun8i);
	sun8i->cachectl.fd = NOFD;

	// Get RX/TX resources (memory for mbufs pointers, TX/RX descriptors memory)
	sun8i->txb.dma = sun8i->rxb.dma = -1;
	sun8i->txb.cnt = CONFIG_TX_DESCR_NUM;
	sun8i->rxb.cnt = CONFIG_RX_DESCR_NUM;
	sun8i->txb.size = sun8i->txb.cnt * sizeof(sun8i_desc_t);
	sun8i->rxb.size = sun8i->rxb.cnt * sizeof(sun8i_desc_t);
	sun8i->txb.tmem_fd = sun8i->rxb.tmem_fd = NOFD;
	sun8i->txb.prot_flags = sun8i->rxb.prot_flags = PROT_READ | PROT_WRITE | PROT_NOCACHE;
	sun8i->txb.map_flags = sun8i->rxb.map_flags = MAP_SHARED | MAP_PHYS | MAP_ANON;

	sun8i->txb.area = mmap(NULL, sun8i->txb.size, sun8i->txb.prot_flags, sun8i->txb.map_flags, sun8i->txb.tmem_fd, 0);
	if (sun8i->txb.area == MAP_FAILED)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "Unable to mmap tx descriptors");
		return (errno);
	}

	if (mem_offset(sun8i->txb.area, sun8i->txb.tmem_fd, sun8i->txb.size, &sun8i->txb.dma, NULL) == -1)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d mem_offset(): %s errno (%d)", __func__, __LINE__, strerror(errno), errno);
		return (errno);
	}

	if ((sun8i->txb.mbufs_set = malloc(sun8i->txb.cnt * sizeof(struct mbuf *), M_DEVBUF, M_NOWAIT)) == NULL)
	{
		return errno;
	}

	sun8i->rxb.area = mmap(NULL, sun8i->rxb.size, sun8i->rxb.prot_flags, sun8i->rxb.map_flags, sun8i->rxb.tmem_fd, 0);
	if (sun8i->rxb.area == MAP_FAILED)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "Unable to mmap tx descriptors");
		return (errno);
	}
	if (mem_offset(sun8i->rxb.area, sun8i->rxb.tmem_fd, sun8i->rxb.size, &sun8i->rxb.dma, NULL) == -1)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d mem_offset(): %s errno (%d)", __func__, __LINE__, strerror(errno), errno);
		return (errno);
	}

	if ((sun8i->rxb.mbufs_set = malloc(sun8i->rxb.cnt * sizeof(struct mbuf *), M_DEVBUF, M_NOWAIT)) == NULL)
	{
		return errno;
	}

	// Initialize cache sync API
	if (cache_init(0, &sun8i->cachectl, NULL) == -1)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "cache_init error %s (%d)", strerror(errno), errno);
		return (errno);
	}

	// Set up process events thread priority (io-pkt get events from interrupt handler *_isr)
	if ((err = interrupt_entry_init(&sun8i->inter, 0, NULL, IRUPT_PRIO_DEFAULT)) != EOK)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d interrupt_entry_init error (%d) ", __func__, __LINE__, err);
		return (err);
	}

	fill_random_mac(mac_addr);
	sun8i_write_mac_addr(sun8i, mac_addr);
	memcpy(sun8i->cfg.current_address, mac_addr, ETHER_ADDR_LEN);
	memcpy(sun8i->cfg.permanent_address, mac_addr, ETHER_ADDR_LEN);

	if_attach(ifp);
	ether_ifattach(ifp, sun8i->cfg.current_address);

	sun8i->sd_hook = shutdownhook_establish(sun8i_shutdown, sun8i);

	return EOK;
}

int sun8i_init(struct ifnet *ifp)
{
	int					i = 0;
	int					rc = 0;
	uint32_t				val = 0;
	struct nw_work_thread			*wtp = WTP;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)ifp->if_softc;
	struct _iopkt_self			*iopkt = sun8i->iopkt;
	sun8i_desc_t				*desc = NULL;
	struct mbuf				*m = NULL;

	// Set non-active state under send-mutex
	NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	ifp->if_flags_tx |= IFF_OACTIVE;
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);

	sun8i_stop(ifp, 0);

	// Clean EMAC registers
	out32(sun8i->emac_base + EMAC_RX_CTL1, 0);
	out32(sun8i->emac_base + EMAC_TX_CTL1, 0);
	out32(sun8i->emac_base + EMAC_RX_DMA_DESC, 0);
	out32(sun8i->emac_base + EMAC_TX_DMA_DESC, 0);
	out32(sun8i->emac_base + EMAC_INT_EN, 0);
	out32(sun8i->emac_base + EMAC_INT_STA, 0x1FFFFFF);

	// Start MDI
	rc = MDI_PowerupPhy(sun8i->mdi, sun8i->cfg.phy_addr);
	if (rc != MDI_SUCCESS)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "PowerupPhy failed");
		return (-1);
	}

	callout_msec(&sun8i->mii_callout, 3 * 1000, sun8i_MDI_MonitorPhy, sun8i);

	MDI_AutoNegotiate(sun8i->mdi, sun8i->cfg.phy_addr, NoWait);

	rc = MDI_EnableMonitor(sun8i->mdi, 1);
	if (rc != MDI_SUCCESS)
	{
		slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d: MDI_EnableMonitor returned %d", __func__, __LINE__, rc);
		return (-1);
	}

	// Set up TX-descriptors (clean status, size, addr, set next for create a queue of descriptors)
	for (i = 0; i < sun8i->txb.cnt; i++)
	{
		desc = SUN8I_TX_DESC(sun8i, i);
		desc->status = 0;
		desc->addr = 0;
		desc->size = 0;
		desc->next = sun8i->txb.dma + ((i + 1) % sun8i->txb.cnt) * sizeof(sun8i_desc_t);
		sun8i->txb.mbufs_set[i] = NULL;
	}
	sun8i->txb.head = 0;
	sun8i->txb.tail = 0;
	sun8i->txb.free_cnt = sun8i->txb.cnt;
	memset(&sun8i->txb.err, 0, sizeof(sun8i->txb.err));

	// Set up RX-descriptors (set owner status, size of max frame size, get mbuf for addr, set next for create a queue of descriptors)
	for (i = 0; i < sun8i->rxb.cnt; i++)
	{
		m = m_getcl_wtp(M_DONTWAIT, MT_DATA, M_PKTHDR, wtp);
		if (m == NULL)
		{
			slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d m_getcl_wtp(): don't get mbuf", __func__, __LINE__);
			return -1;
		}

		sun8i->rxb.mbufs_set[i] = m;
		desc = SUN8I_RX_DESC(sun8i, i);
		desc->addr = mbuf_phys(m);
		desc->size = CONFIG_ETH_RXSIZE;
		desc->status = RX_DESC_CTL;
		desc->next = sun8i->rxb.dma + ((i + 1) % sun8i->rxb.cnt) * sizeof(sun8i_desc_t);

		CACHE_INVAL(&sun8i->cachectl, m->m_data, mbuf_phys(m), m->m_ext.ext_size);
	}
	sun8i->rxb.head = 0;
	sun8i->rxb.tail = 0;
	sun8i->rxb.free_cnt = sun8i->rxb.cnt;
	memset(&sun8i->rxb.err, 0, sizeof(sun8i->rxb.err));

	// Set up start TX/RX descriptors address
	out32(sun8i->emac_base + EMAC_TX_DMA_DESC, sun8i->txb.dma);
	out32(sun8i->emac_base + EMAC_RX_DMA_DESC, sun8i->rxb.dma);

	// Set up TX/RX DMA FIFO starts transmission after get full frame
	val = in32(sun8i->emac_base + EMAC_TX_CTL1);
	val |= EMAC_TX_CTL1_TX_MD;
	out32(sun8i->emac_base + EMAC_TX_CTL1, val);
	val = in32(sun8i->emac_base + EMAC_RX_CTL1);
	val |= EMAC_RX_CTL1_RX_MD;
	out32(sun8i->emac_base + EMAC_RX_CTL1, val);

	// Set up burst len for TX/RX DMA transfer
	out32(sun8i->emac_base + EMAC_CTL1, EMAC_CTL1_BURST_SIZE(8));

	// Start RX/TX DMA
	val = in32(sun8i->emac_base + EMAC_RX_CTL1);
	val |= (EMAC_RX_CTL1_RX_DMA_EN);
	out32(sun8i->emac_base + EMAC_RX_CTL1, val);
	val = in32(sun8i->emac_base + EMAC_TX_CTL1);
	val |= EMAC_TX_CTL1_TX_DMA_EN;
	out32(sun8i->emac_base + EMAC_TX_CTL1, val);

	// Start receiver, transmitter
	val = in32(sun8i->emac_base + EMAC_RX_CTL0);
	val |= EMAC_RX_CTL0_RX_EN;
	out32(sun8i->emac_base + EMAC_RX_CTL0, val);
	val = in32(sun8i->emac_base + EMAC_TX_CTL0);
	val |= EMAC_TX_CTL0_TX_EN;
	out32(sun8i->emac_base + EMAC_TX_CTL0, val);

	if (sun8i->iid == -1)
	{
		if ((rc = InterruptAttach_r(sun8i->cfg.irq[0], sun8i_isr, sun8i, sizeof(sun8i_dev_t), _NTO_INTR_FLAGS_TRK_MSK)) < 0)
		{
			rc = -rc;
			slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d InterruptAttach_r error (%d) ", __func__, __LINE__, rc);
			return (rc);
		}
		sun8i->iid = rc;
	}

	sun8i_enable_interrupt(sun8i);
	out32(sun8i->emac_base + EMAC_INT_STA, 0x1FFFFFF);

	sun8i->linkup = 0;

	NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	ifp->if_flags_tx |= IFF_RUNNING;
	ifp->if_flags_tx &= ~IFF_OACTIVE;
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	ifp->if_flags |= IFF_RUNNING;

	return EOK;
}

void sun8i_stop(struct ifnet *ifp, int disable)
{
	int					i = 0;
	int					rc = 0;
	uint32_t				val = 0;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)ifp->if_softc;
	struct _iopkt_self			*iopkt = sun8i->iopkt;
	struct nw_work_thread			*wtp = WTP;
	sun8i_desc_t				*desc = NULL;

	if_link_state_change(ifp, LINK_STATE_DOWN);
	sun8i->cfg.flags |= NIC_FLAG_LINK_DOWN;
	sun8i->linkup = 0;
	ifp->if_flags &= ~IFF_RUNNING;
	NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	ifp->if_flags_tx &= ~IFF_RUNNING;
	ifp->if_flags_tx |= IFF_OACTIVE;
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);

	callout_stop(&sun8i->mii_callout);
	if (disable)
	{
		slogf(_SLOGC_NETWORK, _SLOG_INFO, "disable interrupt");
		if (sun8i->iid != -1)
		{
			InterruptDetach(sun8i->iid);
			sun8i->iid = -1;
		}
	}

	sun8i_disable_interrupt(sun8i);

	rc = MDI_PowerdownPhy(sun8i->mdi, sun8i->cfg.phy_addr);
	if (rc != MDI_SUCCESS)
	{
		slogf(_SLOGC_NETWORK, _SLOG_ERROR, "PowerdownPhy failed");
	}

	val = in32(sun8i->emac_base + EMAC_RX_CTL1);
	val &= ~(EMAC_RX_CTL1_RX_DMA_EN);
	out32(sun8i->emac_base + EMAC_RX_CTL1, val);

	val = in32(sun8i->emac_base + EMAC_TX_CTL1);
	val &= ~EMAC_TX_CTL1_TX_DMA_EN;
	out32(sun8i->emac_base + EMAC_TX_CTL1, val);

	val = in32(sun8i->emac_base + EMAC_RX_CTL0);
	val &= ~EMAC_RX_CTL0_RX_EN;
	out32(sun8i->emac_base + EMAC_RX_CTL0, val);

	val = in32(sun8i->emac_base + EMAC_TX_CTL0);
	val &= ~EMAC_TX_CTL0_TX_EN;
	out32(sun8i->emac_base + EMAC_TX_CTL0, val);

	for (i = 0; i < sun8i->rxb.cnt; i++)
	{
		desc = SUN8I_DESC(&sun8i->rxb, i);
		desc->addr = 0;
		desc->status = 0;

		if (!sun8i->rxb.mbufs_set[i])
		{
			m_freem(sun8i->rxb.mbufs_set[i]);
			sun8i->rxb.mbufs_set[i] = NULL;
		}
	}

	for (i = 0; i < sun8i->txb.cnt; i++)
	{
		desc = SUN8I_DESC(&sun8i->txb, i);
		desc->addr = 0;
		desc->status = 0;
		desc->size = 0;

		if (!sun8i->txb.mbufs_set[i])
		{
			m_freem(sun8i->txb.mbufs_set[i]);
			sun8i->txb.mbufs_set[i] = NULL;
		}
	}
}

void sun8i_reap(sun8i_dev_t *sun8i)
{
	int					cnt = 0;
	struct mbuf				*m = NULL;
	sun8i_desc_t				*desc = NULL;
	desc_buffer_t				*txb = &sun8i->txb;

	cnt = txb->cnt;
	if (txb->free_cnt == txb->cnt)
	{
		return;
	}

	while (cnt)
	{
		desc = SUN8I_DESC(txb, txb->tail);

		if (desc->status & ERROR_STATUS_MASK)
		{
			txb->err.header = (desc->status & TX_HEADER_ERR) ? txb->err.header: txb->err.header + 1;
			txb->err.length = (desc->status & TX_LENGHT_ERR) ? txb->err.length : txb->err.length + 1;
			txb->err.payload = (desc->status & TX_PAYLOAD_ERR) ? txb->err.payload : txb->err.payload + 1;
			txb->err.crs = (desc->status & TX_CRS_ERR) ? txb->err.crs : txb->err.crs + 1;
			txb->err.coll0 = (desc->status & TX_COL_ERR_0) ? txb->err.coll0 : txb->err.coll0 + 1;
			txb->err.coll1 = (desc->status & TX_COL_ERR_1) ? txb->err.coll1 : txb->err.coll1 + 1;
			txb->err.coll_cnt += TX_COL_CNT(desc->status);
			txb->err.defer = (desc->status & TX_DEFER_ERR) ? txb->err.defer : txb->err.defer + 1;
			txb->err.underflow = (desc->status & TX_UNDERFLOW_ERR) ? txb->err.underflow : txb->err.underflow + 1;
		}

		if (desc->status & TX_DESC_CTL)
		{
			break;
		}

		desc->status = 0;
		desc->addr = 0;
		desc->size = 0;

		if ((m = txb->mbufs_set[txb->tail]) != NULL)
		{
			m_freem(m);
			txb->mbufs_set[txb->tail] = NULL;
		}

		txb->tail = (txb->tail + 1) % txb->cnt;
		++txb->free_cnt;

		if (!txb->free_cnt)
		{
			break;
		}
		--cnt;
	}
}

void sun8i_start(struct ifnet *ifp)
{
	int					num_frag = 0;
	int					end = 0;
	uint32_t				val = 0;
	struct nw_work_thread			*wtp = WTP;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)ifp->if_softc;
	struct _iopkt_self			*iopkt = sun8i->iopkt;
	struct mbuf				*m0 = NULL;
	struct mbuf				*m = NULL;
	sun8i_desc_t				*desc = NULL;
	desc_buffer_t				*txb = &sun8i->txb;

	if (((ifp->if_flags_tx & IFF_RUNNING) == 0) || (!sun8i->linkup))
	{
		goto end;
	}

	ifp->if_flags_tx |= IFF_OACTIVE;
	sun8i->start_running = 1;

	for (;;)
	{
		if (txb->free_cnt < CONFIG_TX_DESCR_NUM / 2)
		{
			sun8i_reap(sun8i);
		}

		IFQ_POLL(&ifp->if_snd, m0);
		if (m0 == NULL)
		{
			break;
		}
		for (num_frag = 0, m = m0; m; m = m->m_next)
		{
			if (m->m_len > 0)
			{
				++num_frag;
			}
		}

		if ((num_frag) > txb->free_cnt)
		{
			sun8i_reap(sun8i);
			if ((num_frag + 2) > txb->free_cnt)
			{
				ifp->if_flags_tx |= IFF_OACTIVE;
				sun8i->start_running = 0;
				NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);

				return;
			}
		}

		IFQ_DEQUEUE(&ifp->if_snd, m0);
		if (m0 == NULL)
		{
			break;
		}

		for (m = m0; m; m = m->m_next)
		{
			if (!m->m_len)
			{
				continue;
			}

			desc = SUN8I_DESC(txb, sun8i->txb.head);
			desc->addr = mbuf_phys(m);
			CACHE_FLUSH(&sun8i->cachectl, m->m_data, desc->addr, m->m_len);
			desc->size = (m == m0) ? FIR_DESC : 0;
			desc->size |= m->m_len | TX_INT_CTL;
			desc->status |= TX_DESC_CTL;

			end = txb->head;
			txb->head = (txb->head + 1) % txb->cnt;
			--txb->free_cnt;
		}
		desc->size |= LAST_DESC;
		txb->mbufs_set[end] = m0;

		val = in32(sun8i->emac_base + EMAC_TX_CTL1);
		val |= (EMAC_TX_CTL1_TX_DMA_EN | EMAC_TX_CTL1_TX_DMA_START);
		out32(sun8i->emac_base + EMAC_TX_CTL1, val);
	}

end:
	ifp->if_flags_tx &= ~IFF_OACTIVE;
	sun8i->start_running = 0;
	NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
}

int sun8i_ioctl(struct ifnet *ifp, unsigned long cmd, caddr_t data)
{
	int					error = 0;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)ifp->if_softc;
	struct ifdrv_com			*ifdc = NULL;
	struct drvcom_stats			*dstp = NULL;
	struct drvcom_config			*dcfgp = NULL;
	struct ifreq				*ifr = NULL;

	switch (cmd)
	{
		case SIOCGDRVCOM:
			ifdc = (struct ifdrv_com *)data;
			switch (ifdc->ifdc_cmd)
			{
				case	DRVCOM_CONFIG:
					dcfgp = (struct drvcom_config *)ifdc;

					if (ifdc->ifdc_len != sizeof(nic_config_t))
					{
						error = EINVAL;
						break;
					}
					memcpy(&dcfgp->dcom_config, &sun8i->cfg, sizeof(sun8i->cfg));

					break;
				case	DRVCOM_STATS:
					dstp = (struct drvcom_stats *)ifdc;

					if (ifdc->ifdc_len != sizeof(nic_stats_t))
					{
						error = EINVAL;
						break;
					}
					memcpy(&dstp->dcom_stats, &sun8i->stats, sizeof(sun8i->stats));

					break;
				default:
					error = ENOTTY;
			}

			break;
		case SIOCSIFMEDIA:
		case SIOCGIFMEDIA:
			ifr = (struct ifreq *)data;
			error = ifmedia_ioctl(ifp, ifr, &sun8i->bsd_mii.mii_media, cmd);

			break;
			/*
		case SIOCSDRVSPEC:
			break;
			*/
		default:
			error = ether_ioctl(ifp, cmd, data);
			if (error == ENETRESET)
			{
				if (error == ENETRESET)
				{
					if (ifp->if_init != NULL)
					{
						ifp->if_init(ifp);
					}
					error = 0;
				}
				break;
			}

			break;
	}
	
	return error;
}

int sun8i_detach(struct device *dev, int flags __attribute__((unused)))
{
	uint32_t				val = 0;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)dev;
	struct ifnet				*ifp = &sun8i->ec.ec_if;

	ether_ifdetach(ifp);
	if_detach(ifp);

	sun8i_stop(ifp, 1);

	shutdownhook_disestablish(sun8i->sd_hook);

	if (sun8i->iid != -1)
	{
		InterruptDetach(sun8i->iid);
		sun8i->iid = -1;
	}
	cache_fini(&sun8i->cachectl);

	interrupt_entry_remove(&sun8i->inter, NULL);

	free(sun8i->txb.mbufs_set, M_DEVBUF);
	free(sun8i->rxb.mbufs_set, M_DEVBUF);
	munmap(sun8i->txb.area, sun8i->txb.size);
	munmap(sun8i->rxb.area, sun8i->rxb.size);

	val = in32(sun8i->clk_base + BUS_CLK_GATING_REG0);
	val |= BUS_CLK_GATING_REG0_EMAC_GATING;
	out32(sun8i->clk_base + BUS_CLK_GATING_REG0, val);

	munmap_device_io(sun8i->syscon_base, AW_H3_SYS_CTRL_SIZE);
	munmap_device_io(sun8i->clk_base, AW_H3_CCU_SIZE);
	munmap_device_io(sun8i->emac_base, SUN8I_EMAC_SIZE);

	return EOK;
}

void sun8i_shutdown(void *arg)
{
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;

	sun8i_stop(&sun8i->ec.ec_if, 1);
}

const struct sigevent *sun8i_isr(void *arg, int iid)
{
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;
	struct _iopkt_inter			*ient = &sun8i->inter;

	sun8i->irq_src = in32(sun8i->emac_base + EMAC_INT_STA);
	out32(sun8i->emac_base + EMAC_INT_STA, sun8i->irq_src);

	sun8i->iid = iid;

	return interrupt_queue(sun8i->iopkt, ient);
}

void sun8i_alloc_mbufs(sun8i_dev_t *sun8i)
{
	struct mbuf				*m = NULL;
	struct nw_work_thread			*wtp = WTP;
	sun8i_desc_t				*desc = NULL;
	desc_buffer_t				*rxb = &sun8i->rxb;

	while (rxb->cnt - rxb->free_cnt)
	{
		m = m_getcl_wtp(M_DONTWAIT, MT_DATA, M_PKTHDR, wtp);
		if (m == NULL)
		{
			slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d m_getcl_wtp(): don't get mbuf", __func__, __LINE__);
			return;
		}

		desc = SUN8I_DESC(rxb, rxb->tail);
		rxb->mbufs_set[rxb->tail] = m;
		desc->addr = mbuf_phys(m);
		CACHE_INVAL(&sun8i->cachectl, m->m_data, mbuf_phys(m), m->m_ext.ext_size);
		desc->status = RX_DESC_CTL;

		rxb->tail = (rxb->tail + 1) % rxb->cnt;
		rxb->free_cnt++;
	}
}

void sun8i_receive(sun8i_dev_t *sun8i)
{
	uint32_t				len = 0;
	uint32_t				total_pkts = 0;
	uint32_t				rx_len = 0;
	off_t					phys = 0;
	struct mbuf				*rm = NULL;
	struct mbuf				*mbuf_head = NULL;
	struct mbuf				**mbuf_tail = &mbuf_head;
	sun8i_desc_t				*desc = NULL;
	desc_buffer_t				*rxb = &sun8i->rxb;
	struct ifnet				*ifp = &sun8i->ec.ec_if;

	desc = SUN8I_DESC(rxb, rxb->head);

	while (!(desc->status & RX_DESC_CTL))
	{
		if (rxb->cnt - rxb->free_cnt > CONFIG_RX_DESCR_NUM / 2)
		{
			sun8i_alloc_mbufs(sun8i);
		}

		len = RX_DESC_FRM_SIZE(desc->status);
		if (!len)
		{
			continue;
		}

		desc->addr = 0;

		rm = rxb->mbufs_set[rxb->head];
		rm->m_len = len;
		rm->m_pkthdr.rcvif = ifp;

		// Collect mbufs queue
		rxb->free_cnt--;
		*mbuf_tail = rm;
		mbuf_tail = &rm->m_next;
		rx_len += len;

		if (!(desc->status & RX_LAST_DESC))
		{
			goto next_mbuf;
		}

		rm = mbuf_head;
		rm->m_pkthdr.len = rx_len;
		rx_len = 0;
		mbuf_head = NULL;
		mbuf_tail = &mbuf_head;
		rm->m_flags |= M_HASFCS;
		phys = pool_phys(rm->m_data, rm->m_ext.ext_page);
		CACHE_INVAL(&sun8i->cachectl, rm->m_data, phys, rm->m_ext.ext_size);

#if NBPFILTER > 0
		// Pass this up to any BPF listeners.
		if (ifp->if_bpf)
		{
			bpf_mtap(ifp->if_bpf, rm);
		}
#endif

		total_pkts++;
		ifp->if_ipackets++;
		(*ifp->if_input)(ifp, rm);

next_mbuf:
		rxb->head = (rxb->head + 1) % rxb->cnt;
		desc = SUN8I_DESC(rxb, rxb->head);
	}

}

int sun8i_xmit_enable_interrupt(void *arg)
{
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;
	struct nw_work_thread			*wtp = WTP;
	struct ifnet				*ifp = &sun8i->ec.ec_if;

	NW_SIGLOCK_P(&ifp->if_snd_ex, sun8i->iopkt, wtp);
	sun8i_start(ifp);

	return 0;
}


int sun8i_process_interrupt(void *arg, struct nw_work_thread *wtp)
{
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;
	struct ifnet				*ifp = &sun8i->ec.ec_if;
	struct _iopkt_self			*iopkt = sun8i->iopkt;

	while (1)
	{
		if (sun8i->irq_src & EMAC_INT_RX_INT)
		{
			sun8i_receive(sun8i);
		}

		if (sun8i->irq_src & EMAC_INT_TX_INT)
		{
			sun8i_xmit_enable_interrupt(sun8i);
		}

		if (sun8i->irq_src & EMAC_INT_TX_UNDERFLOW_INT_EN)
		{
			slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d TX_UNDERFLOW", __func__, __LINE__);
		}

		if (sun8i->irq_src & EMAC_INT_RX_UNDERFLOW_INT_EN)
		{
			slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d RX_UNDERFLOW", __func__, __LINE__);
		}

		sun8i->irq_src = in32(sun8i->emac_base + EMAC_INT_STA);
		out32(sun8i->emac_base + EMAC_INT_STA, sun8i->irq_src);

		if (!(sun8i->irq_src & EMAC_INT_EN_MASK))
		{
			break;
		}
	}
	if (!sun8i->start_running)
	{
		NW_SIGLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
		if (sun8i->txb.free_cnt < CONFIG_TX_DESCR_NUM / 2)
		{
			sun8i_reap(sun8i);
		}
		ifp->if_flags_tx &= ~IFF_OACTIVE;
		NW_SIGUNLOCK_P(&ifp->if_snd_ex, iopkt, wtp);
	}

	return 1;
}

int sun8i_disable_interrupt(void *arg)
{
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;

	out32(sun8i->emac_base + EMAC_INT_EN, 0);

	return 1;
}

int sun8i_enable_interrupt(void *arg)
{
	uint32_t				val = 0;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;

	val = in32(sun8i->emac_base + EMAC_INT_EN);
	val = EMAC_INT_EN_MASK;
	out32(sun8i->emac_base + EMAC_INT_EN, val);

	return 1;
}

void bsd_mii_mediastatus(struct ifnet *ifp, struct ifmediareq *ifmr)
{
	sun8i_dev_t				*sun8i = ifp->if_softc;

	sun8i->bsd_mii.mii_media_active = IFM_ETHER;
	sun8i->bsd_mii.mii_media_status = IFM_AVALID;

	if (sun8i->linkup)
	{
		sun8i->bsd_mii.mii_media_status |= IFM_ACTIVE;

		switch(sun8i->cfg.media_rate)
		{
			case 1000*10:
				sun8i->bsd_mii.mii_media_active |= IFM_10_T;
				break;

			case 1000*100:
				sun8i->bsd_mii.mii_media_active |= IFM_100_TX;
				break;

			default:
				sun8i->bsd_mii.mii_media_active |= IFM_NONE;
				break;
		}

		if (sun8i->cfg.duplex)
		{
			sun8i->bsd_mii.mii_media_active |= IFM_FDX;
		}
	}
	else
	{
		sun8i->bsd_mii.mii_media_active |= IFM_NONE;
	}

	ifmr->ifm_status = sun8i->bsd_mii.mii_media_status;
	ifmr->ifm_active = sun8i->bsd_mii.mii_media_active;
}

int bsd_mii_mediachange(struct ifnet *ifp)
{
	sun8i_dev_t				*sun8i = ifp->if_softc;
	int					old_media_rate = sun8i->cfg.media_rate;
	int					old_duplex = sun8i->cfg.duplex;
	struct ifmedia				*ifm = &sun8i->bsd_mii.mii_media;
	int					user_duplex = ifm->ifm_media & IFM_FDX ? 1 : 0;
	int					user_media = ifm->ifm_media & IFM_TMASK;
	int					media = 0;
	nic_config_t				*cfg = &sun8i->cfg;

	if (!(ifp->if_flags & IFF_UP))
	{
		slogf(_SLOGC_NETWORK, _SLOG_WARNING, "%s(): sun8i interface isn't up, ioctl ignored", __func__);
		return 0;
	}

	if (!(ifm->ifm_media & IFM_ETHER))
	{
		slogf(_SLOGC_NETWORK, _SLOG_WARNING, "%s(): sun8i interface - bad media: 0x%X", __func__, ifm->ifm_media);
		return 0;
	}

	switch (user_media)
	{
		case IFM_AUTO:
			cfg->media_rate = -1;
			cfg->duplex = -1;
			media = IFM_ETHER | IFM_AUTO;
			break;

		case IFM_NONE:
			cfg->media_rate = 0;
			cfg->duplex = 0;
			ifmedia_set(&sun8i->bsd_mii.mii_media, IFM_ETHER | IFM_NONE);
			cfg->flags |= NIC_FLAG_LINK_DOWN;
			if_link_state_change(ifp, LINK_STATE_DOWN);

			return 0;

		case IFM_10_T:
			cfg->media_rate = 10 * 1000;
			media = IFM_ETHER | IFM_10_T;
			break;

		case IFM_100_TX:
			cfg->media_rate = 100 * 1000;
			media = IFM_ETHER | IFM_100_TX;
			break;

		default:
			slogf(_SLOGC_NETWORK, _SLOG_WARNING, "%s(): sun8i interface - unknown media: 0x%X", __func__, user_media);
			return 0;
	}

	cfg->duplex = user_duplex;
	if (user_duplex && user_duplex != -1)
	{
		media |= IFM_FDX;
	}

	media |= 0;

	if ((cfg->media_rate != old_media_rate) ||
		(cfg->duplex != old_duplex) ||
		(cfg->flags & NIC_FLAG_LINK_DOWN))
	{

		cfg->flags |= NIC_FLAG_LINK_DOWN;
		if_link_state_change(ifp, LINK_STATE_DOWN);
		ifp->if_init(ifp);
	}

	return 0;
}

void bsd_mii_initmedia(sun8i_dev_t *sun8i)
{
	nic_config_t				*cfg = &sun8i->cfg;
	struct ifmedia				*ifm = &sun8i->bsd_mii.mii_media;

	sun8i->bsd_mii.mii_ifp = &sun8i->ec.ec_if;

	ifmedia_init(&sun8i->bsd_mii.mii_media, IFM_IMASK, bsd_mii_mediachange, bsd_mii_mediastatus);

	ifmedia_add(&sun8i->bsd_mii.mii_media, IFM_ETHER | IFM_AUTO, 0, NULL);

	switch(cfg->media_rate)
	{
		case -1:
			ifm->ifm_media = IFM_ETHER | IFM_AUTO;
			break;
		case 100 * 1000:
			ifm->ifm_media = IFM_ETHER | IFM_100_TX | IFM_FDX;
			break;
		case 1000 * 1000:
			ifm->ifm_media = IFM_ETHER | IFM_1000_T | IFM_FDX;
			break;
		default:
			slogf(_SLOGC_NETWORK, _SLOG_WARNING, "%s(): unknown initial media(%d), forcing none",  __func__, cfg->media_rate);
			break;
	}

	ifmedia_set(ifm, ifm->ifm_media);
}

void bsd_mii_finimedia(sun8i_dev_t *sun8i)
{
	ifmedia_delete_instance(&sun8i->bsd_mii.mii_media, IFM_INST_ANY);
}

static int sun8i_mdio_busy_wait(sun8i_dev_t *sun8i)
{
	uint64_t				 cycles_max = (~0ULL) ^ ((1ULL << CLOCKCYCLES_INCR_BIT) - 1);
	uint64_t				 cycles_init = ClockCycles();
	uint64_t				 curr;
	uint64_t				 cycles_tm = SYSPAGE_ENTRY(qtime)->cycles_per_sec * 3;

	do
	{
		curr = ClockCycles();
		if (in32(sun8i->emac_base + EMAC_MII_CMD) & MDIO_CMD_MII_BUSY)
		{
			mem_barrier();
		}
		else
		{
			return 0;
		}

		if (cycles_init > curr)
		{
			cycles_init = 0;
			cycles_tm -= cycles_max - cycles_init;
		}
	} while (curr - cycles_init < cycles_tm);

	return -EBUSY;
}

short unsigned int sun8i_mdio_read(void *arg, unsigned char phy_addr, unsigned char phy_reg)
{
	int					data = 0;
	uint32_t				cmd = 0;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;

	cmd = MDIO_CMD_MII_BUSY;
	cmd |= (phy_reg << MDIO_CMD_MII_PHY_REG_ADDR_SHIFT);
	cmd |= (phy_addr << MDIO_CMD_MII_PHY_ADDR_SHIFT);
	cmd |= MDIO_CMD_MDC_DIV_RATIO_M_128;

	if (sun8i_mdio_busy_wait(sun8i))
	{
		return EBUSY;
	}

	out32(sun8i->emac_base + EMAC_MII_DATA, 0);
	out32(sun8i->emac_base + EMAC_MII_CMD, cmd);

	if (sun8i_mdio_busy_wait(sun8i))
	{
		return EBUSY;
	}

	data = in32(sun8i->emac_base + EMAC_MII_DATA);

	return data;
}

void sun8i_mdio_write(void *arg, unsigned char phy_addr, unsigned char phy_reg, short unsigned int phy_data)
{
	uint32_t				cmd = 0;
	sun8i_dev_t				*sun8i = (sun8i_dev_t *)arg;

	cmd = MDIO_CMD_MII_BUSY;
	cmd |= MDIO_CMD_MII_WRITE;
	cmd |= (phy_reg << MDIO_CMD_MII_PHY_REG_ADDR_SHIFT);
	cmd |= (phy_addr << MDIO_CMD_MII_PHY_ADDR_SHIFT);
	cmd |= MDIO_CMD_MDC_DIV_RATIO_M_128;

	if (sun8i_mdio_busy_wait(sun8i))
	{
		return;
	}

	out32(sun8i->emac_base + EMAC_MII_DATA, phy_data);
	out32(sun8i->emac_base + EMAC_MII_CMD, cmd);

	sun8i_mdio_busy_wait(sun8i);
}
