/*
 * (c) 2023, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*
 * Startup for OrangePi PC Allwinner H3 (ARMv7 Cortex-A7 QuadCore)
 */

#include "startup.h"
#include "board.h"
#include "aw_h3_startup.h"

const struct callout_slot callouts[] = {
	{ CALLOUT_SLOT(reboot, _awh3) },
};

/* The USB-Serial is on UART_0 */
const struct debug_device debug_devices[] = {
    {   "8250",
        {   "0x01C28000^2.0.24000000.16",   // UART0, use the baud rate set by boot loader
        },
        init_omap,
        put_omap,
        {   &display_char_8250,
            &poll_key_8250,
            &break_detect_8250,
        }
    },
};


/*
 * main()
 *  Startup program executing out of RAM
 *
 * 1. It gathers information about the system and places it in a structure
 *    called the system page. The kernel references this structure to
 *    determine everything it needs to know about the system. This structure
 *    is also available to user programs (read only if protection is on)
 *    via _syspage->.
 *
 * 2. It (optionally) turns on the MMU and starts the next program
 *    in the image file system.
 */
int
main(int argc, char **argv, char **envv)
{
    int	opt;
	

    while ((opt = getopt(argc, argv, COMMON_OPTIONS_STRING)) != -1) {
        handle_common_option(opt);
    }


    /*
     * Initialize debugging output
     */
    select_debug(debug_devices, sizeof(debug_devices));
    kprintf("Welcome to KPDA Neutrino startup on OrangePi PC Allwinner H3 (ARMv7 Cortex-A7 QuadCore)\n");
    kprintf("Initializing Debug Port...\n");


    add_callout_array(callouts, sizeof(callouts));


    /*
     * Collect information on all free RAM in the system
     */
    init_raminfo();


    /*
     * Remove RAM used by modules in the image
     */
    alloc_ram(shdr->ram_paddr, shdr->ram_size, 1);


    /*
	 * Initialize SMP
	 */
    aw_h3_spinlock_init();
    aw_h3_timer_init();
    init_smp();


    /*
	 * Initialize MMU
	 */
    if (shdr->flags1 & STARTUP_HDR_FLAGS1_VIRTUAL)
    {
        init_mmu();
        kprintf("Initializing MMU...\n");
    }


    /* Initialize the Interrupts related Information */
    init_intrinfo();


    /* Initialize the Timer related information */
    init_qtime();


    /* Init L2 Cache Controller */
    init_cacheattr();


    /* Initialize the CPU related information */
    init_cpuinfo();


    /* Initialize the Hwinfo section of the Syspage */
    init_hwinfo();


    add_typed_string(_CS_MACHINE, "OrangePi PC");

    /*
     * Load bootstrap executables in the image file system and Initialise
     * various syspage pointers. This must be the _last_ initialisation done
     * before transferring control to the next program.
     */
    init_system_private();


    /*
     * This is handy for debugging a new version of the startup program.
     * Commenting this line out will save a great deal of code.
     */
    print_syspage();

    kprintf("Jumping to KPDA\n");

    return 0;
}
