/*
 * (c) 2010-2020, SWD Embedded Systems Limited, http://www.kpda.ru
 */

/*
 * $QNXLicenseC:
 * Copyright 2008, QNX Software Systems. 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"). You 
 * may not reproduce, modify or distribute this software except in 
 * compliance with the License. You may obtain a copy of the License 
 * at: http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as 
 * contributors under the License or as licensors under other terms.  
 * Please review this entire file for other proprietary rights or license 
 * notices, as well as the QNX Development Suite License Guide at 
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */

#include "startup.h"

#ifdef SUPPORT_CMP_LZO
#	include <lzo1x.h>
#endif

#ifdef SUPPORT_CMP_UCL
#	include <ucl/ucl.h>
#endif

#ifdef SUPPORT_CMP_LZ4
#	include <lz4.h>
#endif

#ifdef SUPPORT_CMP_ZSTD
#	define ZSTD_STATIC_LINKING_ONLY
#	include <zstd.h>
extern ZSTD_DCtx* ZSTD_createDCtx_static(void);
#endif

#ifdef SUPPORT_CMP_LZO
void
uncompress_lzo(uint8_t *dst, uint8_t *src) {
	unsigned	len;
	lzo_uint	out_len;
	int			status;

	if(lzo_init() != LZO_E_OK) {
		crash("init fail");
	}
	for(;;) {
		len = (*src++ << 8);
		len += (*src++);
		if(len == 0) break;
		status = lzo1x_decompress(src, len, dst, &out_len, NULL);
		if(status != LZO_E_OK) {
			crash("fail");
		}
		dst += out_len;
		src += len;
	}
}
#endif

#ifdef SUPPORT_CMP_UCL
void
uncompress_ucl(uint8_t *dst, uint8_t *src) {
	unsigned	len;
	ucl_uint	out_len;
	int			status;

	for(;;) {
		len = (*src++ << 8);
		len += (*src++);
		if(len == 0) break;
		status = ucl_nrv2b_decompress_8(src, len, dst, &out_len, NULL);
		if(status != 0) {
			crash("fail");
		}
		dst += out_len;
		src += len;
	}
}
#endif

#ifdef SUPPORT_CMP_LZ4
void
uncompress_lz4(uint8_t *dst, uint8_t *src) {
	unsigned	len;
	int			status;

	for (;;)
	{
		len = (*src++ << 8);
		len += (*src++);
		if (len == 0)
			break;
		status = LZ4_decompress_safe((const char *)src, (char *)dst, len, LZ4_COMPRESSBOUND(0x10000));
		if (status < 0)
		{
			crash("fail");
		}
		dst += status;
		src += len;
	}
}
#endif

#ifdef SUPPORT_CMP_ZSTD
void
uncompress_zstd(uint8_t *dst, uint8_t *src) {
	unsigned	len;
	int			status;
	ZSTD_DCtx *dctx = dctx = ZSTD_createDCtx_static();

	ZSTD_decompressBegin(dctx);

	for (;;)
	{
		len = (*src++ << 8);
		len += (*src++);
		if (len == 0)
			break;
		status = ZSTD_decompressBlock(dctx, dst, ZSTD_COMPRESSBOUND(0x10000), src, len);
		if (status < 0)
		{
			crash("fail");
		}
		dst += status;
		src += len;
	}
}
#endif

void
uncompress(int type, paddr32_t dst_paddr, paddr32_t src_paddr) {
	uint8_t		*dst;
	uint8_t		*src;

	if (debug_flag > 0) kprintf("decompressing...");

	// We assume that the destination is in the one-to-one mapping area.
	dst = MAKE_1TO1_PTR(dst_paddr);
	src = startup_memory_map(shdr->stored_size - shdr->startup_size, src_paddr, PROT_READ);
	switch(type) {
#ifdef SUPPORT_CMP_ZLIB
	case STARTUP_HDR_FLAGS1_COMPRESS_ZLIB:
		uncompress_zlib(dst, 0, src, 0,
				(uint8_t *)shdr->image_paddr + shdr->startup_size + shdr->imagefs_size);
		break;
#endif
#ifdef SUPPORT_CMP_LZO
	case STARTUP_HDR_FLAGS1_COMPRESS_LZO:
		uncompress_lzo(dst, src);
		break;
#endif
#ifdef SUPPORT_CMP_UCL
	case STARTUP_HDR_FLAGS1_COMPRESS_UCL:
		uncompress_ucl(dst, src);
		break;
#endif
#ifdef SUPPORT_CMP_LZ4
	case STARTUP_HDR_FLAGS1_COMPRESS_LZ4:
		uncompress_lz4(dst, src);
		break;
#endif
#ifdef SUPPORT_CMP_ZSTD
	case STARTUP_HDR_FLAGS1_COMPRESS_ZSTD:
		uncompress_zstd(dst, src);
		break;
#endif
	default:
		crash(
			"unsupported compression type (use one of "
#ifdef SUPPORT_CMP_ZLIB
			STARTUP_HDR_FLAGS1_COMPRESS_ZLIB_NAME ", "
#endif
#ifdef SUPPORT_CMP_LZO
			STARTUP_HDR_FLAGS1_COMPRESS_LZO_NAME ", "
#endif
#ifdef SUPPORT_CMP_UCL
			STARTUP_HDR_FLAGS1_COMPRESS_UCL_NAME ", "
#endif
#ifdef SUPPORT_CMP_LZ4
			STARTUP_HDR_FLAGS1_COMPRESS_LZ4_NAME ", "
#endif
#ifdef SUPPORT_CMP_ZSTD
			STARTUP_HDR_FLAGS1_COMPRESS_ZSTD_NAME ", "
#endif
			STARTUP_HDR_FLAGS1_COMPRESS_NONE_NAME ")"
		);
	}
	startup_memory_unmap(src);
}
